/* compare to t_noEC_constrateIncoporate.pl.
The name of simpleRecord refers to the record of visited nodes. 
*/


:- dynamic reaction_state/4, metConcentration/3.

:-['topCla'].
%:-['counting.pl'].
%:-['counting_includeLength'].
:-['b_noCycle'].


%set(depthLimit,s(s(s(s(0))))).
%set(depthLimit,s(s(s(s(s(s(0))))))). %PianoNumDepthLimit):- numMap(6,PianoNumDepthLimit).
%set(depthLimit,s(0)). peanoMinus1(s(X),X).
%set(depthLimit,s(s(0))). peanoMinus1(s(s(X)),X).
%set(depthLimit,s(s(s(0)))). peanoMinus1(s(s(s(X))),X).
set(depthLimit,s(s(s(s(0))))). peanoMinus1(s(s(s(s(X)))),X).

%set(depthLimit,s(s(s(s(s(s(s(0)))))))). % depth 7
%set(depthLimit,s(s(s(s(s(0)))))).
%set(depthLimit,s(s(s(s(s(s(s(s(s(s(0))))))))))).
set(etoB,yes).




% greater than
peanoGth(X,X).
peanoGth(s(Y),X):-peanoGth(Y,X).


/*********** using other examples ***********/
/*concentration(DepthLimit,Tr,Tr_previous,VisitedNodes,X,Change,Time):-	
	set(depthLimit,DepthLimit0),
	DepthLimit \== DepthLimit0,
	ex(EI,concentration(X,Change,Time),1),!,
	Tr=[[met-{X,Change,Time}]|Tr_previous]. % take out the hypothesis about metConcentration if it exist in the examples
	%write('%***using this example***'), write({EI,X,Change}),nl.
	%NewVisitedNodes=[X|VisitedNodes], % since it's the last one, so it doesn't matter in terms of format
	%write('--- one candidate Hypothesis: '),portray_clause(VisitedNodes),pathwayDetails(VisitedNodes).
*/

concentration(DepthLimit,Tr,Tr,VisitedNodes,X,Change,Time):- 
	set(depthLimit,DepthLimit0),
	DepthLimit \== DepthLimit0,
	threeStatesOpposite(Change,Opp1,Opp2),
	ex(EI,concentration(X,Measured,Time),1),
	(Measured=Opp1;Measured=Opp2),
	%write('!!! this metabolite is measured to be '), write({X,Measured}),nl,
	!,
	fail.


threeStatesOpposite(up,down,no_change).
threeStatesOpposite(down,up,no_change).
threeStatesOpposite(no_change,up,down).





/***************************** Regulation Rules ******************************/
/****************** Direct Effect ******************/

/*---------- Blocked/Slower down ----------*/
concentration(s(DepthLimit),TrSoFar,Tr,VisitedNodes,PID,down,Time):-
	produced_by(PID,ReactionID,'IMPD'),
	reaction_state(TrSoFar,Tr,ReactionID,enzymeLimiting,cataDecreased,Time).

concentration(s(CurrentDepth),TrSoFar,Tr,VisitedNodes,SID,up,Time):-				
	set(depthLimit,DepthLimit),peanoMinus1(DepthLimit,DepthLimit1),peanoGth(CurrentDepth,DepthLimit1),%write('succeed'),
	consumed_by(SID,ReactionID,'IMPD'),	
	reaction_state(TrSoFar,Tr,ReactionID,enzymeLimiting,cataDecreased,Time).


/*---------- Speeded Up ----------*/
concentration(s(DepthLimit),TrSoFar,Tr,VisitedNodes,PID,up,Time):-
	produced_by(PID,ReactionID,'IMPD'),
	reaction_state(TrSoFar,Tr,ReactionID,enzymeLimiting,cataIncreased,Time).

concentration(s(CurrentDepth),TrSoFar,Tr,VisitedNodes,SID,down,Time):-					
	set(depthLimit,DepthLimit),peanoMinus1(DepthLimit,DepthLimit1),peanoGth(CurrentDepth,DepthLimit1),%write('succeed'),
	consumed_by(SID,ReactionID,'IMPD'),	
	reaction_state(TrSoFar,Tr,ReactionID,enzymeLimiting,cataIncreased,Time).



/*---------- Catalytically No change ----------*/
concentration(s(DepthLimit),TrSoFar,Tr,VisitedNodes,PID,no_change,Time):-
	produced_by(PID,ReactionID,'IMPD'),
	reaction_state(TrSoFar,Tr,ReactionID,enzymeLimiting,cataNoChange,Time).

concentration(s(CurrentDepth),TrSoFar,Tr,VisitedNodes,SID,no_change,Time):-	
	set(depthLimit,DepthLimit),peanoMinus1(DepthLimit,DepthLimit1),peanoGth(CurrentDepth,DepthLimit1),%write('succeed'),
	consumed_by(SID,ReactionID,'IMPD'),	
	%produced_by(SID,ReactionID_Any,'IMPD'), % complement to the next one, just ensure that there is no redundant %% no matter there is sth to produce or 
	reaction_state(TrSoFar,Tr,ReactionID,enzymeLimiting,cataNoChange,Time).


% This rule doesn't largely increased the candidates, because it requites not produced_by any reactions.
concentration(s(CurrentDepth),TrSoFar,Tr,VisitedNodes,SID,no_change,Time):-					
	set(depthLimit,DepthLimit),peanoMinus1(DepthLimit,DepthLimit1),peanoGth(CurrentDepth,DepthLimit1),%write('succeed'),
	consumed_by(SID,ReactionID,'IMPD'),	
	\+produced_by(SID,ReactionID_Any,'IMPD'),
	reaction_state(TrSoFar,Tr,ReactionID,enzymeLimiting,cataDecreased,Time). 



/****************** Indirect Effect (Substrate Limiting) ******************/
/*------------------------- Substrate Up------------------------*/
concentration(s(DepthLimit),TrSoFar,Tr,VisitedNodes,PID,up,Time):-				
	produced_by(PID,ReactionID,'IMPD'),
	reaction_state(TrSoFar,Tr1,ReactionID,substrateLimiting,sUp,Time),
	consumed_by(SID,ReactionID,'IMPD'), \+member(SID,VisitedNodes),%\+partMember(SID,VisitedNodes),
	Tr2=[[met-{SID,up,Time}]|Tr1],
	concentration(DepthLimit,Tr2,Tr,[SID|VisitedNodes],SID,up,Time).


/*------------------------- Substrate Down ------------------------*/
concentration(s(DepthLimit),TrSoFar,Tr,VisitedNodes,PID,down,Time):-				
	produced_by(PID,ReactionID,'IMPD'),
	reaction_state(TrSoFar,Tr1,ReactionID,substrateLimiting,sDown,Time),
	consumed_by(SID,ReactionID,'IMPD'), \+member(SID,VisitedNodes), %\+partMember(SID,VisitedNodes),
	Tr2=[[met-{SID,down,Time}]|Tr1],
	concentration(DepthLimit,Tr2,Tr,[SID|VisitedNodes],SID,down,Time).



/*
concentration(s(DepthLimit),TrSoFar,Tr,VisitedNodes,PID,no_change,Time):-				
	produced_by(PID,ReactionID,'IMPD'),
	reaction_state(TrSoFar,Tr1,ReactionID,substrateLimiting,cataDecreased,Time),
	consumed_by(SID,ReactionID,'IMPD'),\+partMember(SID,VisitedNodes),
	concentration(DepthLimit,Tr1,Tr,[{ReactionID,cataDecreased}-PID|VisitedNodes],SID,up,Time).

concentration(s(DepthLimit),TrSoFar,Tr,VisitedNodes,PID,no_change,Time):-				
	produced_by(PID,ReactionID,'IMPD'),	
	reaction_state(TrSoFar,Tr1,ReactionID,substrateLimiting,cataIncreased,Time),
	consumed_by(SID,ReactionID,'IMPD'),\+partMember(SID,VisitedNodes),
	concentration(DepthLimit,Tr1,Tr,[{ReactionID,cataIncreased}-PID|VisitedNodes],SID,down,Time).
*/
% ??? What about substrate no_change?







/*************** Abduce reaction_state	****************/
/* Simple version -- no incorporation of integrity constraint
reaction_state(TrSoFar,Tr,EC_Number,State):-
	Tr=[[reaction_state-{EC_Number,State}]|TrSoFar].
*/


% incorporate integrity constraint
% 1. one enzyme not to be hypothesized to several states: (1) the existing B; (2) the same 
% 2. the gene expression data. 

% This is not suitable when there are alternative similar rules 


% I don't think I really need Sign to be recorded, but to ensure to be used together with deriving hypotheses


reaction_state(TrSoFar,Tr,ReactionID,LimitingType,State,Time):-
	(set(ge_data_available,yes)->
		constraintCheck(ReactionID,LimitingType,State,Time);
		Foo=0
	),
	(appeared(ReactionID,TrSoFar,LimitingType2,State2)-> 
		State2==State, LimitingType2==LimitingType, % write('not contradiction'),nl,	% this will rarely happen -- unless it is going through a cycle, which is not allowed
		Tr=TrSoFar;
		Tr=[[rs-{ReactionID,LimitingType,State,Time}]|TrSoFar]	% Sign here will be left as a variable
	).

% 

appeared(ReactionID,TrSoFar,LimitingType,State):-	
	member([rs-{ReactionID,LimitingType,State,Time}],TrSoFar).


/*
constraintCheck(ReactionID,LimitingType,HypothesizedState,Time):-
	catalyzed_by_ECclass(ReactionID,EC_Number), %timeEquivenlent(Time,DoseDependantTime),
	concentration_e(EC_Number,GEChange,Degree,Time),!,
	(LimitingType==enzymeLimiting ->
		proportional(HypothesizedState,GEChange); % enzyme limiting, require it to be proportional -- if not annotated.
		Foo=1	%
		%nonContradict(HypothesizedState,GEChange) % substrate limiting
	).	
constraintCheck(ReactionID,substrateLimiting,HypothesizedState,Time). % always pass the test if they are not annotated.
*/



%constraintCheck(ReactionID,substrateLimiting,HypothesizedState,Time):-!. % always pass the test if they are substractLimiting.
% if enzyme limiting, then always require GE to support. -- Here the assumption is, those without GE are very expensive (in terms of DL), so they won't be selected in the end, so not to generate from the first instance
% if the there is one reaction without EC annotated, while is the shared point of more than one metabolites, then you may miss some good solution. 
constraintCheck(ReactionID,enzymeLimiting,cataNoChange,Time):-!. % hypothesising no change will always pass
constraintCheck(ReactionID,enzymeLimiting,HypothesizedState,Time):- 
	catalyzed_by_ECclass(ReactionID,EC_Number), 
	concentration_e(EC_Number,GEChange,Degree,Time),!,
	proportional(HypothesizedState,GEChange).
constraintCheck(ReactionID,enzymeLimiting,HypothesizedState,Time). % always passed if no GE annotated

/*constraintCheck(ReactionID,substrateLimiting,HypothesizedState,Time):-
      	catalyzed_by_ECclass(ReactionID,EC_Number),
	concentration_e(EC_Number,GEChange,Degree,Time),!,
	nonContradict(HypothesizedState,GEChange).
*/
constraintCheck(ReactionID,substrateLimiting,HypothesizedState,Time):-
      	catalyzed_by_ECclass(ReactionID,EC_Number),
	concentration_e(EC_Number,GEChange,Degree,Time),!,
	(nonContradict(HypothesizedState,GEChange)->
		Foo=0; % do nothing and pass
			getEarlierTimes(Time,EarlierTimes),
	maplist(checkSubstrateLimitingContradict({EC_Number,HypothesizedState}),EarlierTimes,EarlierTimes_Consistent0),flatten(EarlierTimes_Consistent0,EarlierTimes_Consistent),
		length(EarlierTimes_Consistent,ConsistentNum),ConsistentNum >=1	% go to earlier time point \+keep this contradict -- in the early time point it is consiste
	).
constraintCheck(ReactionID,substrateLimiting,HypothesizedState,Time).

/* These already exist in the file of 'scoringScheme.pl' (in the same folder of t.pl)
proportional(cataIncreased,up).
proportional(cataDecreased,down).
proportional(cataNoChange,noChange).

nonContradict(sUp,GEChange):- GEChange\==down.
nonContradict(sDown,GEChange):- GEChange\==up.
nonContradict(sNoChange,GEChange):- GEChange=='noChange'.
*/


getEarlierTimes('Day1',[]).
getEarlierTimes('Day3',['Day1']).
getEarlierTimes('Day7',['Day1','Day3']).
getEarlierTimes('Day14',['Day1','Day3','Day7']).


      
checkSubstrateLimitingContradict({EC_Number,HypothesizedState},Time,ConsistentSign):-
	concentration_e(EC_Number,GEChange,Degree,Time),
	(nonContradict(HypothesizedState,GEChange)->
		ConsistentSign=[Time]; % do nothing and pass
		ConsistentSign=[]
	).
      


