% These rules are similar to that in ILP2010, but add activation
% The rules are divided according to enzyme limiting or substrate limiting, which is decided by the availability of corresponding gene expression data. 



%enzyme_state(EC_Number,no_change):- \+enzyme_state(EC_Number,inhibited).

concentration(X,Change):-	% using other examples
	ex(EI,concentration(X,Change),1),!.
% won't be used in reasoning, but used in T for excluding those useless
concentration(X,Change):- 
	neg(Change,OppositeChange),
	ex(EI,concentration(X,OppositeChange),1),%write('!!! you are looking for sth inconsistent'),
	!,
	fail.


% link reactions to their EC_Number 
catalyzed_by_ECclass(ReactionID,EC_Number):-
	setof(EC_Num,
		EnzID^(catalyzed_by(ReactionID,EnzID,'IMPD'),
		part_of_catalyzing_class(EnzID,EnzClassID,'IMPD'),
		concept_accession(EnzClassID,'EC',EC_Num,false)),
	EC_Nums),
	member(EC_Number,EC_Nums).



/****************** Direct Effect ******************/
% 12,15,18 enzyme inhibited
concentration(PID,down):-				%concept_name(P,PID,true),	
	produced_by(PID,ReactionID,'IMPD'),	%reaction(ReactionID,'','LycoCyc','IMPD'),
	catalyzed_by_ECclass(ReactionID,EC_Number),
	\+concentration_enzClass(EC_Number,'up'),
	enzyme_state(EC_Number,inhibited).

/*----------Change in Substrates ----------*/
% Substrates are consumed more or less with the change of reaction speed 

% enzyme inhibited
concentration(SID,up):-				%concept_name(S,SID,true),	
	consumed_by(SID,ReactionID,'IMPD'),	%reaction(ReactionID,'','LycoCyc','IMPD'),
	catalyzed_by_ECclass(ReactionID,EC_Number),
	\+concentration_enzClass(EC_Number,'up'),
	enzyme_state(EC_Number,inhibited).


/****************** Indirect Effect ******************/
% enzyme no_change -- S up
concentration(PID,up):-				%concept_name(P,PID,true),	
	produced_by(PID,ReactionID,'IMPD'),	%reaction(ReactionID,'','LycoCyc','IMPD'),
	catalyzed_by_ECclass(ReactionID,EC_Number), 
	enzyme_state(EC_Number,no_change),
	consumed_by(SID,ReactionID,'IMPD'), concentration(SID,up).

% enzyme no_change -- S down
concentration(PID,down):-				%concept_name(P,PID,true),	
	produced_by(PID,ReactionID,'IMPD'),	%reaction(ReactionID,'','LycoCyc','IMPD'),
	catalyzed_by_ECclass(ReactionID,EC_Number),
	enzyme_state(EC_Number,no_change),
	consumed_by(SID,ReactionID,'IMPD'), concentration(SID,down).



/************ Integrity Constraint **************/
integrityConstraint:-
enzyme_state(EC_Number,State1),enzyme_state(EC_Number,State2),State1\==State2. %,write('Integrity Constraint is Disobeyed'),nl.


/****************** Simple direct rules ******************/
%This integrity constraint can be put into top theory directly.
/* this can be transform into integrity constraint 
enzyme_state(EC_Number,activated):- concentration_enzClass(EC_Number,up).
enzyme_state(EC_Number,no_change):- concentration_enzClass(EC_Number,no_change).
enzyme_state(EC_Number,inhibited):- concentration_enzClass(EC_Number,down).
*/

set(specificScore,yes).

scoreSpecific(T,Coverage,Score):-
	tCompScore(T,TComplex),
	%Score is TComplex+Coverage.
	CoverageScore is 2*Coverage,
	Score is TComplex+CoverageScore.


tCompScore([],0).
tCompScore([ClaI|T],NTScore):-
	tCompScore(T,TScore), 
	nl,write(ClaI),nl,
	(ClaI=[enzyme_state-{EC_Number,no_change}]->
		NTScore=TScore;
		claInterpreter(ClaI,Cla),
		length(Cla,HL),
		NTScore is TScore+HL
	).	


